

set CustomGUI 0
set Customcfgfile ""

set WriteIF 0x70     ;#   User Register 0
set ReadIF 0x74      ;#   User Register 1

set clock_frequency      100.0

set decimation_factor    4
set shaping_time         6.0
set trapezoid_width      4.4
set peak_separation      8.5
set decay_time_constant  50.0
set peak_sample_time     7.2
set baseline_update_time 2
set baseline_average     4
set baseline_threshold   16384
set mode                 0
set tfa_shaping_time     120.0
set tfa_decay_time       50000.0  


for {set reg 16} {$reg <= 31} {incr reg} {set MWDr$reg 0}
set MWDr36 0     ;#  TFA
set MWDr37 0     ;#  TFA

for {set i 0} {$i < 13} {incr i} {set MWDp$i 0}

set UserReg {}
for {set reg 16} {$reg <= 31} {incr reg} {lappend UserReg $reg}
lappend UserReg 36
lappend UserReg 37

if {$UserGUIVersion == 2} {     ;#   MWD firmware
set Reg0x00 0     ;#  control register
set Reg0x20 0     ;#  CFD Threshold
set Reg0x21 0     ;#  Veto Control
set Reg0x26 0     ;#  Hit pattern trigger mask
set Reg0x30 0     ;#  Trace length 
set Reg0x31 0     ;#  PreTrigger Position
set Reg0x3a 0
set Reg0x3b 0
set Reg0x3c 0
set Reg0x3d 0
set Reg0x70 0    ;# TDRI Control
set Reg0x71 0    ;# Veto Coincidence window
set Reg0x72 0    ;# GPIO Control
set Reg0x73 0    ;# VETO delay
set Reg0x74 0    ;# VETO delay
set Reg0x75 0    ;# VETO delay
set Reg0x76 0    ;# VETO delay
set Reg0x77 0    ;# VETO delay
set Reg0x78 0    ;# VETO delay
set Reg0x79 0    ;# VETO delay
set Reg0x7A 0    ;# VETO delay

set Reg0x80 0
set Reg0x81 0
set Reg0x83 0

set Reg0x90 0     ;#  PSA registers
set Reg0x91 0
set Reg0x92 0
set Reg0x93 0
set Reg0x94 0
set Reg0x95 0
set Reg0x96 0
set Reg0x97 0

set RegList {}
foreach item [list 00 20 21 26 30 31 3a 3b 3c 3d 70 71 72 73 74 75 76 77 78 79 7A] {
   lappend RegList 0x$item
}
foreach item [list 90 91 92 93 94 95 96 97] {
   lappend RegList 0x$item
}

}

if {$UserGUIVersion == 1} {  
set Reg0x00 0
set RegList {0x00}
}  

set user10 0
set user11 0
set user12 0
set user90 0
set user91 0
set user92 0

if {$UserGUIVersion == 1} {    

    for {set i 0} {$i <= 31} {incr i} {set RegMask($i) 0}
}

if {$UserGUIVersion == 2} {     ;#   MWD firmware

      for {set i 0} {$i <= 31} {incr i} {set RegMask($i) 0x0000ffff}
      set RegMask(36) 0x0000ffff
      set RegMask(37) 0x0000ffff

      set RegMask(16) 255
      set RegMask(17) 255
      set RegMask(20) 255
      set RegMask(21) 255
      set RegMask(24) 255
      set RegMask(25) 31
      set RegMask(26) 3
}


proc ReadUserReg {chan reg} {
    global boardID

        set regvl [expr ($chan << 28) + ($reg << 16)]        ;#    build data for register
#insert-debug "Reading [format 0x%08x $regvl] from board Index $boardID"
        set value [ReadVHSADCUserRegisters $boardID $regvl] 
#insert-debug "Read $value from board Index $boardID"

    return $value
}

proc WriteUserReg {chan reg vl} {
    global boardID

        set regvl [expr ($chan << 28) + ($reg << 16) + $vl]        ;#    build data for register
#insert-debug "Writing [format 0x%08x $regvl] to board Index $boardID"
        VHSADCAccessClient__WriteUserRegs $boardID $regvl   
}

proc WriteUserRegs {ID registers} {

#insert-debug "Writing $registers to board Index $ID"
#     foreach reg $registers {insert-debug "Writing [format 0x%8x $reg] to board Index $ID"}
        VHSADCAccessClient__WriteUserRegs $ID $registers        ;#    write to list of user registers
}

proc ReadUserRegs {ID registers} {

#insert-debug "Reading $registers from board Index $ID"
        set values [ReadVHSADCUserRegisters $ID $registers]        ;#   read from list of user register
#insert-debug "Read $values from board Index $ID"
     return $values
}

proc do_pause  {} {inform sys "click-special 23"}
proc do_resume {} {inform sys "click-special 24"}

proc click-select-chan {vl} {
    global chan
    global NumChan boardID CrateNum BoardNum ChanNum

    if {$vl >= $NumChan($boardID)} {
         midas-warning "Channel not supported by FPGA firmware"
         set chan 0
    } else {
         set chan $vl
    }

    set ChanNum [expr $chan + 1]

    :modify CHAN-NUM :value $ChanNum

    update_UserGUI
}

proc Read_MWD_Registers {} {
    global boardID chan
    global RegMask
    global UserReg RegList
    global Reg0x80 Reg0x81 Reg0x83

    foreach reg $UserReg {global MWDr$reg}
    foreach reg $RegList {global Reg$reg}

    set registers {}    ;#  build list of required registers

    foreach reg $UserReg {lappend registers  [expr ($chan << 28) + ($reg << 16)]}

    foreach reg $RegList {
        lappend registers  [expr ($chan << 28) + ($reg << 16)]
    }

    lappend registers [expr 0x80 << 16]
    lappend registers [expr 0x81 << 16]
    lappend registers [expr 0x83 << 16]

    set values [ReadUserRegs $boardID $registers]   ;# get all the values

    set index 0

    foreach reg $UserReg {
        set rc [lindex $values $index]; incr index
        set MWDr$reg [expr $rc & $RegMask($reg)]
    }

    foreach reg $RegList {
        set rc [lindex $values $index]; incr index
        set Reg$reg [expr $rc & 0x0000ffff]
    }

    foreach reg {0x80 0x81 0x83} {
        set rc [lindex $values $index]; incr index
        set Reg$reg [expr $rc & 0x0000ffff]
    }
}

proc Set_MWD_Registers {} {
    global UserGUIVersion

    if {$UserGUIVersion == 2} {     ;#   MWD firmware
        update_mwd_parameters
        apply_mwd_parameters
    }

    update_UserGUI   ;# if needed

}

proc click_UserGUI {id vl index} {
    global boardID boardlist chan chanlist NumChan
    global allmodules allchans
    global Settings

# insert-debug "click_UserGUI with $id $vl $index"

    set chan_SAV $chan
    set boardID_SAV $boardID

   switch $id {
     910 -
     911 -
     912 -
     920 -
     921 -
     922 -
     923 {
          :modify [self frame-name].user :busy T
          do_click_UserGUI $chan $id $vl $index
          :modify [self frame-name].user :busy F
          return
     }
     913 {set allchans 0}   ;#  note executed below
   }

    :modify [self frame-name].user :busy T

    if {$allmodules == 0} {
        set Settings {}
        if {$allchans == 0} {
            do_click_UserGUI $chan $id $vl $index
        } else {
            foreach chan $chanlist {
                 if {$chan >= $NumChan($boardID)} {break}
                 Read_MWD_Registers
                 do_click_UserGUI $chan $id $vl $index
            }
        }
        if {$Settings != ""} {
            set z [catch {WriteUserRegs $boardID $Settings} m]
            if {$z != 0} {
                insert "Error writing UserRegs to board Index $boardID (rc=$z; m=$m)"
            }
        }
    } else {
        foreach boardID $boardlist {

            set Settings {}
            if {$allchans == 0} {
                do_click_UserGUI $chan $id $vl $index
            } else {
                foreach chan $chanlist {
                    if {$chan >= $NumChan($boardID)} {break}
                    Read_MWD_Registers
                    do_click_UserGUI $chan $id $vl $index
                }
            }
            if {$Settings != ""} {
                set z [catch {WriteUserRegs $boardID $Settings} m]
                if {$z != 0} {
                    insert "Error writing UserRegs to board Index $boardID (rc=$z; m=$m)"
                }
            }
        }
    }

    set boardID $boardID_SAV
    set chan $chan_SAV

    :modify [self frame-name].user :busy F
    clear-footer [self frame-name].user

    if {$allmodules != 0} {click-select-module [expr $boardID+1]}
    if {$allchans != 0} {click-select-chan $chan} 

    Set_MWD_Registers

}

proc do_click_UserGUI {chan id vl index} {    ;#  only called from click_UserGUI to set parameter
    global UserGUIVersion
    global boardID WriteIF ReadIF
    global user10 user11 user12
    global user90 user91 user92
    global Settings
    global UserReg RegList
    foreach reg $UserReg {global MWDr$reg}
    foreach reg $RegList {global Reg$reg}
    global Reg0x24 Reg0x25

# insert-debug "do_click_UserGUI with $boardID $chan $id $vl $index"

    switch $id {
        910 {set user10 $index}
        911 {set user11 $index}
        912 {set user12 $vl}
        913 {                          ;# click on button USER913   (Write User Register 1)
            WriteUserReg $user10 $user11 $user12
        }
        920 {set user90 $index}
        921 {set user91 $index}
        922 {set user92 $vl}
        923 {                          ;# click on button USER923   (Read User Register 2)
            set rc [ReadUserReg $user90 $user91]
            :modify USER92  :inactive F :value [format 0x%08x $rc] :inactive T
            if {[expr ($rc >> 28) & 15] == $user90 && [expr ($rc >> 16) & 0x0fff] == $user91} {
                :modify USER922 :value [format 0x%04x [expr $rc & 0x0000ffff]]
            } else {
                :modify USER922 :value "******"
            }
        }

        000 {
            set vl [expr ($Reg0x00 & 0xfffe) | $vl]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        001 {
            set vl [expr ($Reg0x00 & 0xfffd) | ($vl << 1)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        002 {
            set vl [expr ($Reg0x00 & 0xfff3) | ($vl << 2)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        004 {
            set vl [expr ($Reg0x00 & 0xffef) | ($vl << 4)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        005 {
            set vl [expr ($Reg0x00 & 0xff1f) | ($vl << 5)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        008 {
            set vl [expr ($Reg0x00 & 0xfcff) | ($vl << 8)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        010 {
            set vl [expr ($Reg0x00 & 0xe3ff) | ($vl << 10)]
            set reg 0x00
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }

        20  -
        21  -
        26  -
        31  {
            set reg 0x$id
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
        }

        30  {
            set reg 0x$id
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $index]
        }

        70  -
        72  {
            set reg 0x$id
            lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr $vl<<1]]
        }

        71D15 {
            set reg 0x71
            lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr ($vl << 15) | ($Reg0x71 & 0x7fff)]]
        }
        71D14 {
            set reg 0x71
            lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr ($vl << 14) | ($Reg0x71 & 0xbfff)]]
        }
        71  {
            set reg 0x71
            lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr ($vl & 0x3fff) | ($Reg0x71 & 0xc000)]]
        }

        VENABLE {
            for {set i 0} {$i < 7} {incr i} {
                set reg 0x[expr 73+$i]
                set VL [expr ([set Reg$reg] & 0x7f) + ((($vl>>$i)&1)<<7)]
                lappend Settings [expr ($chan << 28) + ($reg << 16) + $VL]
            }
            set i 7
            set reg 0x7A
            set VL [expr ([set Reg$reg] & 0x7f) + ((($vl>>$i)&1)<<7)]
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $VL]
        }
        73  -
        74  -
        75  -
        76  -
        77  -
        78  -
        79  -
        7A  {
            set reg 0x$id
            lappend Settings [expr ($chan << 28) + ($reg << 16) + ([set Reg$reg] & 0x80) + (($vl/10) & 0x7f)]
        }

        R0x24 -
        R0x25 -
        R16 -
        R17 -
        R18 -
        R19 -
        R20 -
        R21 -
        R22 -
        R24 -
        R25 -
        R26 {
            set reg [string range $id 1 end]
            lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            if {[expr $Reg0x00 & 1] == 0} {  ;# channel enabled so needs a reset
                set reg 0x00
                lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr $Reg0x00 | 1]]
                lappend Settings [expr ($chan << 28) + ($reg << 16) + $Reg0x00]
            }
        }
        R23 {
            if {$UserGUIVersion == 2} {     ;#   MWD firmware
                set reg 23
                lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr $vl & 0x0000ffff]]
                if {[expr $Reg0x00 & 1] == 0} {  ;# channel enabled so needs a reset
                   set reg 0x00
                   lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr $Reg0x00 | 1]]
                   lappend Settings [expr ($chan << 28) + ($reg << 16) + $Reg0x00]
                }
            }
            if {$UserGUIVersion == 1} { 
                set reg [string range $id 1 end]
                lappend Settings [expr ($chan << 28) + ($reg << 16) + $vl]
            }
        }
    }
}


proc click-clock_frequency  {v} {
    global clock_frequency ADCClockFreq

    if {$ADCClockFreq != "External"} {
        midas-warning "For internal clocks you should set the frequency when loading the FPGA fimrware"
        return
    }

    set clock_frequency $v
    Set_MWD_Registers
}

proc click_trigger {w v} {

    switch $w {
       TRIGMASKOPTS   {
          switch $v {
             0  {set vl 65535}
             1  {set vl 0}
          }
          :modify TRIGMASKOPTS :value -1
          click_UserGUI 26 $vl 0
       }
       TRIGMASKBITS {set vl $v; click_UserGUI 26 $vl 0}
       TRIGMODE     {
          switch $v {
            0  {set b10 0; set b11 1; set b12 0}
            1  {set b10 0; set b11 1; set b12 1}
            2  {set b10 1; set b11 0; set b12 1}
            3  {set b10 1; set b11 1; set b12 1}
          }
          click_UserGUI 010 [expr ($b12*4) + ($b11*2) + $b10] 0
       }
    }
}

proc click_mwd_parameter {args} {
    global clock_frequency decimation_factor shaping_time  trapezoid_width
    global peak_separation decay_time_constant peak_sample_time
    global baseline_average  baseline_update_time baseline_threshold mode
    global tfa_shaping_time 
#   global tfa_decay_time

    for {set i 0} {$i < 13} {incr i} { global MWDp$i }

    set w [string tolower [first $args]]
    set v [second $args]

#insert-debug "mwd_parameters $w $v"

#insert-debug "MWDp0=$MWDp0; MWDp7=$MWDp7"

    set $w $v
    apply_mwd_parameters

#insert-debug "MWDp0=$MWDp0; MWDp7=$MWDp7"

    switch $w {
        mode               {click_UserGUI R26 $MWDp10 0}
        baseline_threshold {click_UserGUI R22 $MWDp6 0}
        baseline_average   {click_UserGUI R20 $MWDp4 0}
        baseline_update_time {click_UserGUI R21 $MWDp5 0}
        peak_sample_time   {click_UserGUI R18 $MWDp2 0}
        decay_time_constant  {click_UserGUI R23 $MWDp7 0}
        peak_separation    {click_UserGUI R19 $MWDp3 0}
        trapezoid_width    {click_UserGUI R17 $MWDp1 0}
        shaping_time       {set p7 $MWDp7; click_UserGUI R16 $MWDp0 0; click_UserGUI R23 $p7 0}
        decimation_factor  {click_UserGUI R25 $MWDp9 0}
        tfa_shaping_time   {set p11 $MWDp11; click_UserGUI R0x25 $MWDp12 0; click_UserGUI R0x24 $p11 0}
        tfa_decay_time     {click_UserGUI R0x24 $MWDp11 0}
    }
}

proc apply_mwd_parameters {} {
    global clock_frequency decimation_factor shaping_time  trapezoid_width
    global peak_separation decay_time_constant peak_sample_time
    global baseline_average  baseline_update_time baseline_threshold mode
    global tfa_shaping_time
    global tfa_decay_time
    global ADCClockFreq

    for {set i 0} {$i < 13} {incr i} { global MWDp$i }

    if {$ADCClockFreq != "External"} {
        if {$clock_frequency != $ADCClockFreq} {set clock_frequency $ADCClockFreq;  update_mwd_parameters}
    }

    set clock_width [expr 1000000000 / ( [set clock_frequency] * 1000000 ) ]
    set ddt [expr $clock_width * 0.001 * $decimation_factor ]
    set dt [expr $clock_width * 0.001 ]
    set Mbydt [expr $shaping_time / $ddt]

    set MWDp0 [expr round(256 - $Mbydt +1)]                             ;# R16
    set MWDp1 [expr 256 - round($trapezoid_width / $ddt ) -1]           ;# R17
    set MWDp2 [expr int($peak_sample_time / $dt)]                       ;# R18
    set MWDp3 [expr int( $peak_separation / $dt ) ]                     ;# R19
    set MWDp4 $baseline_average                                         ;# R20
    set MWDp5 [expr round($baseline_update_time / $ddt)]                ;# R21
    set MWDp6 $baseline_threshold                                       ;# R22

    set z [catch {expr round(16777216 * (1 - exp([expr - $shaping_time / $decay_time_constant])) / $Mbydt)} m]    ;# R23
    if {$z != 0} {midas-warning "Unacceptable data value"; return}
    set MWDp7 $m

    set MWDp9 [expr 32 - $decimation_factor +1]                         ;# R25
    set MWDp10 $mode                                                    ;# R26

    set z [catch {expr round(256-($tfa_shaping_time*1.0/10))} m]
    if {$z != 0} {midas-warning "Unacceptable data value"; return}
    set MWDp12 $m

    set z [catch {expr round(0x01000000 * (1-exp(-$tfa_shaping_time*1.0/$tfa_decay_time))/($tfa_shaping_time/10))} m]
    if {$z != 0} {midas-warning "Unacceptable data value"; return}
    set MWDp11 $m

}

proc update_mwd_parameters {} {
    global clock_frequency decimation_factor shaping_time  trapezoid_width
    global peak_separation decay_time_constant peak_sample_time 
    global baseline_average baseline_update_time baseline_threshold mode
    global tfa_shaping_time 
#   global tfa_decay_time
    global ADCClockFreq

    for {set reg 16} {$reg <= 31} {incr reg} {global MWDr$reg}
    global MWDr36 MWDr37

    if {$ADCClockFreq != "External"} {
        if {$clock_frequency != $ADCClockFreq} {set clock_frequency $ADCClockFreq;  apply_mwd_parameters}
    }

    set clock_width [expr 1000000000 / ( [set clock_frequency] * 1000000 ) ]
    set decimation_factor [expr 33 - $MWDr25]
    set ddt [expr $clock_width * 0.001 * $decimation_factor ]
    set dt [expr $clock_width * 0.001 ]
    set shaping_time [expr (257 - $MWDr16) * $ddt]
    set Mbydt [expr $shaping_time / $ddt]

#    set shaping_time [expr 257 - $MWDr16]
    set trapezoid_width [expr (255 - $MWDr17) * $ddt]
    set peak_sample_time [expr $MWDr18 * $dt]
    set peak_separation [expr $MWDr19 * $dt]
    set baseline_average $MWDr20
    set baseline_update_time [expr $MWDr21 * $ddt]
    set baseline_threshold $MWDr22
    set p1 [expr ($MWDr23*$Mbydt)/16777216]
    set p2 [expr 1/(1-$p1)]
    set decay_time_constant [expr $shaping_time/log($p2)]
#    set decimation_factor [expr 33 - $MWDr25]
    set mode $MWDr26
    set tfa_shaping_time [expr (256-$MWDr37)*10] 
#    set tfa_decay_time  [expr $tfa_shaping_time/log(1.0/(1-($MWDr36*($tfa_shaping_time/10)*1.0)/0x1000000))]

    redisplay_mwd_parameters

}

proc click-applysettings {which} {
    global boardlist chanlist boardID chan NumChan
    global allmodules

   if {$which == 0} {return}

    set-footer "Apply Settings $which"

    :modify [self frame-name].user :busy T

   set Num 0

   foreach ID $boardlist {

       if {$allmodules == 0 && $ID != $boardID} {continue}

       foreach C $chanlist {

          if {$ID == $boardID && $C == $chan} {continue}
          if {$C >= $NumChan($ID)} {break}

          if {[expr ($which >> ($C+1)) & 1] != 1} {continue}

          set boardID_SAV $boardID
          set chan_SAV $chan

          set boardID $ID
          set chan $C

          set Settings {}    ;#  build list of required register settings

          foreach reg [list 0x00] {  ;# channel control register (per channel) 
             global Reg$reg
             lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]
          }
          for {set reg 16} {$reg <= 26} {incr reg} {   ;#  MWD parameters
             global MWDr$reg
             lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set MWDr$reg]]
          }
          for {set reg 36} {$reg <= 37} {incr reg} {   ;#  TFA parameters
             global MWDr$reg
             lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set MWDr$reg]]
          }
          foreach reg [list 0x20 0x21 0x26] {  ;# CFD Threshold; VETO control  (per channel) 
             global Reg$reg
             lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]
          }
          foreach reg [list 0x30 0x31] {   ;# Trace length; PreTrigger Position(per card)
             global Reg$reg
             if {$C == 0} {lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]}
          }
          foreach reg [list 0x70 0x71 0x72 0x73 0x74 0x75 0x76 0x77 0x78 0x79 0x7A] { ;# VETO + GPIO Ctrl (per card)
             global Reg$reg
             if {$C == 0} {lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]}
          }
          foreach reg [list 0x90 0x91 0x92 0x93 0x94 0x95 0x96 0x97] { ;# PSA Ctrl (per card)
             global Reg$reg
             if {$C == 0} {lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]}
          }
          foreach reg [list 0x3a 0x3b 0x3c 0x3d] {   ;# TimeStamp Preload (per card)
             global Reg$reg
             
             if {$C == 0} {lappend Settings  [expr ($chan << 28) + ($reg << 16) + [set Reg$reg]]}
          }

            if {[expr $Reg0x00 & 1] == 0} {  ;# channel enabled so needs a reset
                set reg 0x00
                lappend Settings [expr ($chan << 28) + ($reg << 16) + [expr $Reg0x00 | 1]]
                lappend Settings [expr ($chan << 28) + ($reg << 16) + $Reg0x00]
            }

          WriteUserRegs $boardID $Settings   ;# set all the values

          incr Num [llength $Settings]

          set boardID $boardID_SAV
          set chan $chan_SAV
       }
    }

    click-select-module [expr $boardID+1]
    click-select-chan $chan

    Set_MWD_Registers

    :modify [self frame-name].user :busy F

    set-footer [self frame-name].user "$Num settings updated"
    after 2500 clear-footer [self frame-name].user
}


